from datetime import datetime
import logging
from pathlib import Path
import tkinter
from tkinter import filedialog, messagebox, simpledialog

from packetraven import DEFAULT_CALLSIGNS
from packetraven.connections import APRS_fi, PacketRadio, PacketTextFile, next_available_port
from packetraven.tracks import APRSTrack
from packetraven.utilities import get_logger
from packetraven.writer import write_aprs_packet_tracks

LOGGER = get_logger('packetraven_gui')

DEFAULT_INTERVAL_SECONDS = 5
DESKTOP_PATH = Path('~').expanduser() / 'Desktop'


class PacketRavenGUI:
    def __init__(self):
        self.main_window = tkinter.Tk()
        self.main_window.title('packetraven main')

        self.connections = []

        self.active = False
        self.packet_tracks = {}

        self.frames = {}
        self.elements = {}
        self.last_row = 0

        self.frames['top'] = tkinter.Frame(self.main_window)
        self.frames['top'].pack()

        self.frames['separator'] = tkinter.Frame(height=2, bd=1, relief=tkinter.SUNKEN)
        self.frames['separator'].pack(fill=tkinter.X, padx=5, pady=5)

        self.frames['bottom'] = tkinter.Frame(self.main_window)
        self.frames['bottom'].pack()

        self.__add_entry_box(self.frames['top'], 'port')

        self.__add_entry_box(self.frames['top'], title='log_file', width=45)
        self.elements['log_file'].insert(0, DESKTOP_PATH / f'packetraven_log_{datetime.now():%Y%m%dT%H%M%S}.txt')
        log_file_button = tkinter.Button(self.frames['top'], text='...', command=self.__select_log_file)
        log_file_button.grid(row=self.last_row, column=2)

        self.__add_entry_box(self.frames['top'], title='output_file', width=45)
        self.elements['output_file'].insert(0, DESKTOP_PATH / f'packetraven_output_{datetime.now():%Y%m%dT%H%M%S}.geojson')
        output_file_button = tkinter.Button(self.frames['top'], text='...', command=self.__select_output_file)
        output_file_button.grid(row=self.last_row, column=2)

        self.toggle_text = tkinter.StringVar()
        self.toggle_text.set('Start')
        toggle_button = tkinter.Button(self.frames['top'], textvariable=self.toggle_text, command=self.toggle)
        toggle_button.grid(row=self.last_row + 1, column=1)
        self.last_row += 1

        self.__add_text_box(self.frames['bottom'], title='longitude', units='°')
        self.__add_text_box(self.frames['bottom'], title='latitude', units='°')
        self.__add_text_box(self.frames['bottom'], title='altitude', units='m')
        self.__add_text_box(self.frames['bottom'], title='ground_speed', units='m/s')
        self.__add_text_box(self.frames['bottom'], title='ascent_rate', units='m/s')

        for element in self.frames['bottom'].winfo_children():
            element.configure(state=tkinter.DISABLED)

        try:
            self.serial_port = next_available_port()
            self.replace_text(self.elements['port'], self.serial_port)
        except OSError:
            self.serial_port = None

        self.main_window.mainloop()

    def __add_text_box(self, frame: tkinter.Frame, title: str, units: str = None, row: int = None, entry: bool = False,
                       width: int = 10):
        if row is None:
            row = self.last_row + 1

        column = 0

        element_label = tkinter.Label(frame, text=title)
        element_label.grid(row=row, column=column)

        column += 1

        if entry:
            element = tkinter.Entry(frame, width=width)
        else:
            element = tkinter.Text(frame, width=width, height=1)

        element.grid(row=row, column=column)

        column += 1

        if units is not None:
            units_label = tkinter.Label(frame, text=units)
            units_label.grid(row=row, column=column)

        column += 1

        self.last_row = row

        self.elements[title] = element

    def __add_entry_box(self, frame: tkinter.Frame, title: str, row: int = None, width: int = 10):
        self.__add_text_box(frame, title, row=row, entry=True, width=width)

    def __select_log_file(self):
        filename = Path(self.elements['log_file'].get()).stem
        path = filedialog.asksaveasfilename(title='PacketRaven log location...', initialfile=filename,
                                            defaultextension='.txt', filetypes=[('Text', '*.txt')])

        if path != '':
            self.replace_text(self.elements['log_file'], path)

    def __select_output_file(self):
        filename = Path(self.elements['output_file'].get()).stem
        path = filedialog.asksaveasfilename(title='PacketRaven output location...', initialfile=filename,
                                            defaultextension='.kml', filetypes=[('GeoJSON', '*.geojson'), ('Keyhole Markup Language', '*.kml')])
        if path != '':
            self.replace_text(self.elements['output_file'], path)

    def toggle(self):
        if self.active:
            for connection in self.connections:
                connection.close()

                if type(connection) is PacketRadio:
                    LOGGER.info(f'closing port {connection.location}')

            for element in self.frames['bottom'].winfo_children():
                element.configure(state=tkinter.DISABLED)

            self.toggle_text.set('Start')
            self.active = False
            self.connections = []

            LOGGER.info('stopping')
            logging.shutdown()
        else:
            log_filename = self.elements['log_file'].get()
            get_logger(LOGGER.name, log_filename)
            LOGGER.info('starting')

            try:
                if self.serial_port is None:
                    self.serial_port = self.elements['port'].get()

                    if self.serial_port == '':
                        try:
                            self.serial_port = next_available_port()
                            self.replace_text(self.elements['port'], self.serial_port)
                        except Exception as error:
                            LOGGER.exception(f'{error.__class__.__name__} - {error}')
                            self.serial_port = None

                if self.serial_port is not None:
                    if 'txt' in self.serial_port:
                        try:
                            text_file = PacketTextFile(self.serial_port)
                            self.connections.append(text_file)
                        except Exception as error:
                            LOGGER.exception(f'{error.__class__.__name__} - {error}')
                    else:
                        try:
                            radio = PacketRadio(self.serial_port)
                            self.serial_port = radio.location
                            LOGGER.info(f'opened port {self.serial_port}')
                            self.connections.append(radio)
                        except Exception as error:
                            LOGGER.exception(f'{error.__class__.__name__} - {error}')

                aprs_fi_api_key = simpledialog.askstring('APRS.fi API Key', 'enter API key for https://aprs.fi', parent=self.main_window)

                try:
                    aprs_api = APRS_fi(DEFAULT_CALLSIGNS, api_key=aprs_fi_api_key)
                    LOGGER.info(f'established connection to API')
                    self.connections.append(aprs_api)
                except Exception as error:
                    LOGGER.exception(f'{error.__class__.__name__} - {error}')

                if len(self.connections) == 0:
                    raise ConnectionError('Could not establish connection to APRS packet source.')

                for element in self.frames['bottom'].winfo_children():
                    element.configure(state=tkinter.NORMAL)

                self.toggle_text.set('Stop')
                self.active = True
            except Exception as error:
                messagebox.showerror('PacketRaven Error', f'{error.__class__.__name__} - {error}')
                self.active = False

            self.run()

    def run(self):
        if self.active:
            LOGGER.debug(f'receiving packets from {len(self.connections)} source(s)')

            parsed_packets = []
            for connection in self.connections:
                parsed_packets.extend(connection.packets)

            LOGGER.debug(f'received {len(parsed_packets)} packets')

            if len(parsed_packets) > 0:
                for parsed_packet in parsed_packets:
                    callsign = parsed_packet['callsign']

                    if callsign in self.packet_tracks:
                        if parsed_packet not in self.packet_tracks[callsign]:
                            self.packet_tracks[callsign].append(parsed_packet)
                        else:
                            LOGGER.debug(f'{callsign:8} - received duplicate packet: {parsed_packet}')
                            continue
                    else:
                        self.packet_tracks[callsign] = APRSTrack(callsign, [parsed_packet])
                        LOGGER.debug(f'{callsign:8} - started tracking')

                    LOGGER.info(f'{callsign:8} - received new packet: {parsed_packet}')
                    packet_track = self.packet_tracks[callsign]

                    if 'longitude' in parsed_packet and 'latitude' in parsed_packet:
                        coordinates = packet_track.coordinates[-1]
                        ascent_rate = packet_track.ascent_rate[-1]
                        ground_speed = packet_track.ground_speed[-1]
                        seconds_to_impact = packet_track.seconds_to_impact

                        LOGGER.info(f'{callsign:8} - Ascent rate (m/s): {ascent_rate}')
                        LOGGER.info(f'{callsign:8} - Ground speed (m/s): {ground_speed}')
                        if seconds_to_impact >= 0:
                            LOGGER.info(f'{callsign:8} - Estimated time until landing (s): {seconds_to_impact}')

                        if callsign in DEFAULT_CALLSIGNS:
                            self.replace_text(self.elements['longitude'], coordinates[0])
                            self.replace_text(self.elements['latitude'], coordinates[1])
                            self.replace_text(self.elements['altitude'], coordinates[2])
                            self.replace_text(self.elements['ground_speed'], ground_speed)
                            self.replace_text(self.elements['ascent_rate'], ascent_rate)

                output_filename = self.elements['output_file'].get()
                if output_filename != '':
                    write_aprs_packet_tracks(self.packet_tracks.values(), output_filename)

            if self.active:
                self.main_window.after(DEFAULT_INTERVAL_SECONDS * 1000, self.run)

    @staticmethod
    def replace_text(element, value):
        if type(element) is tkinter.Text:
            start_index = '1.0'
        else:
            start_index = 0

        element.delete(start_index, tkinter.END)
        element.insert(start_index, value)


def main():
    PacketRavenGUI()


if __name__ == '__main__':
    main()



from datetime import datetime, timedelta
import math
from typing import Union

import numpy
from pyproj import CRS, Geod, Transformer

from packetraven.parsing import parse_raw_aprs

DEFAULT_CRS = CRS.from_epsg(4326)


class LocationPacket:
    """ location packet encoding (x, y, z) and time """

    def __init__(self, time: datetime, x: float, y: float, z: float = None, crs: CRS = None):
        self.time = time
        self.coordinates = numpy.array((x, y, z if z is not None else 0))
        self.crs = crs if crs is not None else DEFAULT_CRS

    def distance(self, point: (float, float)) -> float:
        """
        horizontal distance over ellipsoid

        :param point: (x, y) point
        :return: distance in ellipsodal units
        """

        if not isinstance(point, numpy.ndarray):
            point = numpy.array(point)
        coordinates = numpy.stack([self.coordinates[:2], point], axis=0)
        if self.crs.is_projected:
            return numpy.hypot(*numpy.sum(numpy.diff(coordinates, axis=0), axis=0))
        else:
            ellipsoid = self.crs.datum.to_json_dict()['ellipsoid']
            geodetic = Geod(a=ellipsoid['semi_major_axis'], rf=ellipsoid['inverse_flattening'])
            return geodetic.line_length(coordinates[:, 0], coordinates[:, 1])

    def transform_to(self, crs: CRS):
        transformer = Transformer.from_crs(self.crs, crs)
        self.coordinates = transformer.transform(self.coordinates)

    def __sub__(self, other: 'LocationPacket') -> 'LocationPacketDelta':
        """
        Return subtraction of packets in the form of Delta object.

        :param other: location packet
        :return: Delta object
        """

        other_coordinates = Transformer.from_crs(other.crs, self.crs).transform(*other.coordinates) if other.crs != self.crs else other.coordinates

        seconds = (self.time - other.time) / timedelta(seconds=1)
        horizontal_distance = self.distance(other_coordinates[:2])
        vertical_distance = self.coordinates[2] - other_coordinates[2]

        return LocationPacketDelta(seconds, horizontal_distance, vertical_distance)

    def __eq__(self, other: 'LocationPacket') -> bool:
        """
        Whether this packet equals another packet, ignoring datetime (because of possible staggered duplicate packets).

        :param other: packet to compare to this one
        :return: equality
        """

        return numpy.allclose(self.coordinates, other.coordinates)

    def __gt__(self, other: 'LocationPacket') -> bool:
        """
        Whether this packet is after another packet in time.

        :param other: packet to compare to this one
        :return: whether this packet occurred after the other
        """

        return self.time > other.time

    def __lt__(self, other: 'LocationPacket') -> bool:
        """
        Whether this packet is before another packet in time.

        :param other: packet to compare to this one
        :return: whether this packet occurred before the other
        """

        return self.time < other.time

    def __str__(self) -> str:
        return f'{self.time} {self.coordinates}'


class LocationPacketDelta:
    def __init__(self, seconds: float, horizontal_distance: float, vertical_distance: float):
        self.seconds = seconds
        self.vertical_distance = vertical_distance
        self.horizontal_distance = horizontal_distance

    @property
    def distance(self):
        # TODO account for ellipsoid
        return numpy.hypot(self.horizontal_distance, self.vertical_distance)

    @property
    def ascent_rate(self):
        return self.vertical_distance / self.seconds if self.seconds > 0 else math.inf

    @property
    def ground_speed(self):
        return self.horizontal_distance / self.seconds if self.seconds > 0 else math.inf

    def __str__(self) -> str:
        return f'{self.seconds} s, {self.vertical_distance:6.2f} m vertical, {self.horizontal_distance:6.2f} m horizontal'


class APRSLocationPacket(LocationPacket):
    """ APRS packet containing parsed APRS fields, along with location and time """

    def __init__(self, time: datetime, x: float, y: float, z: float, crs: CRS = None, **kwargs):
        """
        APRS packet object from raw packet and given datetime

        :param x: x value
        :param y: y value
        :param z: z value
        :param crs: coordinate reference system
        :param time: time of packet, either as datetime object, seconds since Unix epoch, or ISO format date string.
        """

        super().__init__(time, x, y, z, crs)
        self.parsed_packet = kwargs

    @classmethod
    def from_raw_aprs(cls, raw_aprs: Union[str, bytes, dict], time: datetime = None) -> 'APRSLocationPacket':
        """
        APRS packet object from raw packet and given datetime

        :param raw_aprs: string containing raw packet
        :param time: Time of packet, either as datetime object, seconds since Unix epoch, or ISO format date string.
        """

        # parse packet with metric units
        parsed_packet = parse_raw_aprs(raw_aprs)

        if 'longitude' in parsed_packet and 'latitude' in parsed_packet:
            if time is None:
                if 'timestamp' in parsed_packet:
                    # extract time from Unix epoch
                    time = datetime.fromtimestamp(float(parsed_packet['timestamp']))
                else:
                    # TODO make HABduino add timestamp to packet upon transmission
                    # otherwise default to time the packet was received (now)
                    time = datetime.now()

            return cls(time, parsed_packet['longitude'], parsed_packet['latitude'], parsed_packet['altitude'] if 'altitude' in parsed_packet else None, crs=DEFAULT_CRS, **parsed_packet)
        else:
            raise ValueError(f'Input packet does not contain location data: {raw_aprs}')

    @property
    def callsign(self) -> str:
        return self['callsign']

    def __getitem__(self, field: str):
        if field == 'callsign':
            field = 'from'

        if self.__contains__(field):
            return self.parsed_packet[field]
        else:
            raise KeyError(f'Packet does not contain the field "{field}"')

    def __contains__(self, field: str):
        """
        whether packet contains the given APRS field

        :param field: APRS field name
        :return: whether field exists
        """

        if field == 'callsign':
            field = 'from'

        return field in self.parsed_packet

    def __iter__(self):
        yield from self.parsed_packet

    def __eq__(self, other: 'APRSLocationPacket') -> bool:
        """
        whether this packet equals another packet, including callsign and comment

        :param other: packet to compare to this one
        :return: equality
        """

        return super().__eq__(other) and self.callsign == other.callsign and self['comment'] == other['comment']

    def __str__(self) -> str:
        return f'{self["callsign"]} {super().__str__()} "{self["comment"]}"'

    def __repr__(self) -> str:
        return str(self)



from ast import literal_eval
from datetime import date, datetime
from functools import partial
from typing import Any, Union

import psycopg2
from pyproj import CRS
from shapely import wkb, wkt
from shapely.errors import WKBReadingError, WKTReadingError
from shapely.geometry import LineString, LinearRing, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon, shape as shapely_shape
from shapely.geometry.base import BaseGeometry
from sshtunnel import SSHTunnelForwarder

from packetraven.utilities import get_logger

LOGGER = get_logger('packetraven.connection')

POSTGRES_DEFAULT_PORT = 5432
SSH_DEFAULT_PORT = 22

POSTGRES_TYPES = {
    'NoneType' : 'NULL',
    'bool'     : 'BOOL',
    'float'    : 'REAL',
    'int'      : 'INTEGER',
    'str'      : 'VARCHAR',
    'bytes'    : 'BYTEA',
    'date'     : 'DATE',
    'time'     : 'TIME',
    'datetime' : 'TIMESTAMP',
    'timedelta': 'INTERVAL',
    'list'     : 'VARCHAR[]',
    'dict'     : 'HSTORE',
    'ipaddress': 'INET',
    'Point'    : 'GEOMETRY'
}

DEFAULT_CRS = CRS.from_epsg(4326)
GEOMETRY_TYPES = (Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon)


class InheritedTableError(Exception):
    pass


class DatabaseTable:
    def __init__(self, hostname: str, database: str, table: str, fields: {str: type}, primary_key: str = None, crs: CRS = None, username: str = None, password: str = None, users: [str] = None,
                 **kwargs):
        self.hostname = hostname
        self.database = database
        self.table = table

        self.fields = fields
        self.primary_key = primary_key
        self.crs = crs if crs is not None else DEFAULT_CRS

        self.username = username
        self.password = password
        if users is None:
            users = []

        # parse port from URL
        self.hostname, self.port = split_URL_port(hostname)
        if self.port is None:
            self.port = POSTGRES_DEFAULT_PORT

        connector = partial(psycopg2.connect, database=self.database, user=self.username, password=self.password, port=self.port)
        if 'ssh_hostname' in kwargs or 'ssh_port' in kwargs:
            with SSHTunnelForwarder((kwargs['ssh_hostname'] if 'ssh_hostname' in kwargs else self.hostname,
                                     int(kwargs['ssh_port']) if 'ssh_port' in kwargs else SSH_DEFAULT_PORT),
                                    ssh_username=kwargs['ssh_username'] if 'ssh_username' in kwargs else None,
                                    ssh_password=kwargs['ssh_password'] if 'ssh_password' in kwargs else None,
                                    remote_bind_address=('localhost', self.port)):
                self.connection = connector(host='localhost')
        else:
            self.connection = connector(host=self.hostname)

        with self.connection:
            with self.connection.cursor() as cursor:
                if database_has_table(cursor, self.table):
                    if database_table_is_inherited(cursor, self.table):
                        raise InheritedTableError(f'inheritance of table "{self.database}/{self.table}" will cause unexpected behaviour; aborting')

                    remote_fields = self.remote_fields
                    if list(remote_fields) != list(self.fields):
                        LOGGER.warning(f'schema of existing table "{self.database}/{self.table}" differs from given fields')

                        remote_fields_not_in_local_table = {field: value for field, value in remote_fields.items() if field not in self.fields}
                        if len(remote_fields_not_in_local_table) > 0:
                            LOGGER.warning(f'remote table has {len(remote_fields_not_in_local_table)} fields not in local table: {list(remote_fields_not_in_local_table)}')
                            LOGGER.warning(f'adding {len(remote_fields_not_in_local_table)} fields to local table: {list(remote_fields_not_in_local_table)}')

                            for field, field_type in remote_fields_not_in_local_table.items():
                                previous_field = list(remote_fields)[list(remote_fields).index(field) - 1]

                                local_fields = {}
                                for local_field, local_value in self.fields.items():
                                    local_fields[local_field] = local_value
                                    if local_field == previous_field:
                                        local_fields[field] = field_type

                                self.fields = local_fields

                        local_fields_not_in_remote_table = {field: value for field, value in self.fields.items() if field not in remote_fields}
                        if len(local_fields_not_in_remote_table) > 0:
                            LOGGER.warning(f'local table has {len(local_fields_not_in_remote_table)} fields not in remote table: {list(local_fields_not_in_remote_table)}')
                            LOGGER.warning(f'adding {len(local_fields_not_in_remote_table)} fields to remote table: {list(local_fields_not_in_remote_table)}')

                        if list(remote_fields) != list(self.fields):
                            LOGGER.warning(f'altering schema of "{self.database}/{self.table}"')
                            LOGGER.debug(self.remote_fields)
                            LOGGER.debug(self.fields)

                            copy_table_name = f'old_{self.table}'

                            if database_has_table(cursor, copy_table_name):
                                cursor.execute(f'DROP TABLE {copy_table_name};')

                            cursor.execute(f'ALTER TABLE {self.table} RENAME TO {copy_table_name};')

                            cursor.execute(f'CREATE TABLE {self.table} ({self.schema});')
                            for user in users:
                                cursor.execute(f'GRANT INSERT, SELECT, UPDATE, DELETE ON TABLE public.{self.table} TO {user};')

                            cursor.execute('SELECT column_name FROM information_schema.columns WHERE table_name=%s;', [f'{copy_table_name}'])
                            column_names = [record[0] for record in cursor.fetchall()]

                            cursor.execute(f'INSERT INTO {self.table} ({", ".join(column_names)}) SELECT * FROM {copy_table_name};')

                            cursor.execute(f'DROP TABLE {copy_table_name};')
                else:
                    LOGGER.warning(f'creating remote table "{self.database}/{self.table}"')
                    cursor.execute(f'CREATE TABLE {self.table} ({self.schema});')

                    for user in users:
                        cursor.execute(f'GRANT INSERT, SELECT, UPDATE, DELETE ON TABLE public.{self.table} TO {user};')

    def __getitem__(self, key: Any) -> {str: Any}:
        """
        Query table for the given value of the primary key.

        Parameters
        ----------
        key
            value to query from primary key

        Returns
        -------
        {str: Any}
            dictionary record
        """

        records = self.records_where({self.primary_key: key})

        if len(records) > 1:
            LOGGER.warning(f'found more than one record with primary key "{self.primary_key} = {key}": {records}')

        return records[0]

    def __setitem__(self, key: Any, record: {str: Any}):
        """
        Insert the given record into the table with the given primary key value.

        Parameters
        ----------
        key
            value of primary key at which to insert record
        record
            dictionary record
        """

        record[self.primary_key] = key
        self.insert([record])

    @property
    def records(self) -> [{str: Any}]:
        with self.connection:
            with self.connection.cursor() as cursor:
                cursor.execute(f'SELECT {", ".join(self.fields.keys())} FROM {self.table}')
                records = cursor.fetchall()

        return [parse_record_values(dict(zip(self.fields.keys(), record)), self.fields) for record in records]

    def records_where(self, where: {str: Union[Any, list]}) -> [{str: Any}]:
        """
        records in the table that match the given key-value pairs

        Parameters
        ----------
        where
            dictionary mapping keys to values, with which to match records

        Returns
        -------
        [{str: Any}]
            dictionaries of matching records
        """

        records = []
        try:
            for record in self.records:
                match = True
                for key, value in where.items():
                    if key in record:
                        record_value = record[key]
                        if type(value) in [tuple, list, slice, range]:
                            if record_value not in value:
                                match = False
                        elif value != record_value if value is not None else value is not record_value:
                            match = False
                    else:
                        match = False
                if match:
                    records.append(record)
        except KeyError:
            LOGGER.warning(f'no records found with given constraints: {where}')
        return records

    def insert(self, records: [{str: Any}]):
        """
        Insert the list of records into the table.

        Parameters
        ----------
        records
            dictionary records
        """

        if type(records) is dict:
            records = [records]

        assert all(self.primary_key in record for record in records), f'one or more records does not contain primary key "{self.primary_key}"'

        with self.connection:
            with self.connection.cursor() as cursor:
                for record in records:
                    record_fields_not_in_local_table = [field for field in record if field not in self.fields]
                    if len(record_fields_not_in_local_table) > 0:
                        LOGGER.warning(f'record has {len(record_fields_not_in_local_table)} fields not in the local table that will not be inserted: {record_fields_not_in_local_table}')

                    local_fields_in_record = [field for field in self.fields if field in record]
                    geometry_fields = [field for field in self.geometry_fields if field in record]

                    columns = [field for field in local_fields_in_record if field not in geometry_fields]
                    values = [record[field] for field in local_fields_in_record if field not in geometry_fields]

                    if database_table_has_record(cursor, self.table, record, self.primary_key):
                        record_without_primary_key = {column: value for column, value in zip(columns, values) if column != self.primary_key}

                        if len(record_without_primary_key) > 0:
                            if len(record_without_primary_key) > 1:
                                cursor.execute(f'UPDATE {self.table} SET ({", ".join(record_without_primary_key.keys())}) = %s WHERE {self.primary_key} = %s;',
                                               [tuple(record_without_primary_key.values()), record[self.primary_key]])
                            else:
                                cursor.execute(f'UPDATE {self.table} SET {tuple(record_without_primary_key.keys())[0]} = %s WHERE {self.primary_key} = %s;',
                                               [tuple(record_without_primary_key.values())[0], record[self.primary_key]])
                    else:
                        cursor.execute(f'INSERT INTO {self.table} ({", ".join(columns)}) VALUES %s;', [tuple(values)])

                    if len(geometry_fields) > 0:
                        geometries = {field: record[field] for field in geometry_fields}

                        for field, geometry in geometries.items():
                            cursor.execute(f'UPDATE {self.table} SET {field} = ST_GeomFromWKB(%s::geometry, %s) ' +
                                           f'WHERE {self.primary_key} = %s;', [geometry.wkb, self.crs.to_epsg(), record[self.primary_key]])

    @property
    def schema(self) -> str:
        """ PostGres schema string of local table, given field names and types """

        schema = []
        for field, field_type in self.fields.items():
            dimensions = 0
            while type(field_type) is list:
                field_type = field_type[0]
                dimensions += 1

            field_definition = f'{field} {POSTGRES_TYPES[field_type.__name__]}{"[]" * dimensions}'

            if field == self.primary_key:
                field_definition = f'{field_definition} PRIMARY KEY'

            schema.append(field_definition)

        return ', '.join(schema)

    @property
    def remote_fields(self) -> {str: type}:
        with self.connection:
            with self.connection.cursor() as cursor:
                if database_has_table(cursor, self.table):
                    cursor.execute(f'SELECT column_name, udt_name FROM information_schema.columns WHERE table_name=%s;', [self.table])
                    fields = {field[0]: field[1] for field in cursor.fetchall()}

                    for field, field_type in fields.items():
                        dimensions = field_type.count('_')
                        field_type = field_type.strip('_')

                        field_type = field_type.lower()
                        for python_type, postgres_type in POSTGRES_TYPES.items():
                            if postgres_type.lower() == field_type:
                                field_type = eval(python_type)
                                break
                        else:
                            for python_type, postgres_type in POSTGRES_TYPES.items():
                                if python_type.lower() in field_type:
                                    field_type = eval(python_type)
                                    break
                            else:
                                field_type = str

                        for _ in range(dimensions):
                            field_type = [field_type]
                        fields[field] = field_type

                    return fields

    @property
    def geometry_fields(self) -> {str: type}:
        return {field: field_type for field, field_type in self.fields.items() if field_type in GEOMETRY_TYPES}

    def __contains__(self, key: Any) -> bool:
        with self.connection:
            with self.connection.cursor() as cursor:
                return database_table_has_record(cursor, self.table, {self.primary_key: key})

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.hostname}, {self.database}, {self.table}, {self.fields}, {self.primary_key}, {self.crs}, {self.username}, {self.password})'


def parse_record_values(record: {str: Any}, field_types: {str: type}) -> {str: Any}:
    """
    Parse the values in the given record into their respective field types.

    Parameters
    ----------
    record
        dictionary mapping fields to values
    field_types
        dictionary mapping fields to types

    Returns
    -------
    {str: Any}
        record with values parsed into their respective types
    """

    for field, value in record.items():
        if field in field_types:
            field_type = field_types[field]
            value_type = type(value)

            if value_type is not field_type:
                if field_type is bool:
                    value = bool(value) if value_type is not str else literal_eval(value.capitalize())
                elif field_type is int:
                    value = int(value)
                elif field_type is float:
                    value = float(value)
                elif field_type is str:
                    value = str(value)
                elif value_type is str:
                    if field_type is list:
                        value = literal_eval(value)
                    elif field_type in (date, datetime):
                        value = datetime.strptime(value, '%Y%m%d')
                        if field_type is date:
                            value = value.date()
                    elif field_type in GEOMETRY_TYPES:
                        try:
                            value = wkb.loads(value, hex=True)
                        except WKBReadingError:
                            try:
                                value = wkt.loads(value)
                            except WKTReadingError:
                                try:
                                    value = wkb.loads(value)
                                except TypeError:
                                    value = shapely_shape(literal_eval(value))
                record[field] = value
    return record


def flatten_geometry(geometry: BaseGeometry) -> BaseGeometry:
    geometry_type = type(geometry)

    # strip 3rd dimension
    if 'POLYGON Z' in geometry.wkt:
        polygons = [polygon for polygon in geometry] if geometry_type is MultiPolygon else [geometry]
        for polygon_index, polygon in enumerate(polygons):
            exterior_2d = LinearRing([vertex[:2] for vertex in polygon.exterior.coords])
            interiors_2d = [LinearRing([vertex[:2] for vertex in interior.coords]) for interior in polygon.interiors]
            polygons[polygon_index] = Polygon(exterior_2d, interiors_2d)
        geometry = MultiPolygon(polygons) if geometry_type is MultiPolygon else Polygon(polygons[0])

    if not geometry.is_valid:
        geometry = geometry.buffer(0)
    return geometry


def database_has_table(cursor: psycopg2._psycopg.cursor, table: str) -> bool:
    """
    Whether the given table exists within the given database.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table

    Returns
    -------
    bool
        whether table exists
    """

    cursor.execute(f'SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE table_name=%s);', [table.lower()])
    return cursor.fetchone()[0]


def database_table_has_record(cursor: psycopg2._psycopg.cursor, table: str, record: {str: Any}, primary_key: str = None) -> bool:
    """
    Whether the given record exists within the given table.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table
    record
        dictionary record
    primary_key
        name of primary key

    Returns
    -------
    bool
        whether record exists in table
    """

    if primary_key is None:
        # cursor.execute(f'SELECT 1 FROM information_schema.table_constraints WHERE table_name=\'{table}\' AND constraint_type= \'PRIMARY KEY\';')
        # primary_key_index = cursor.fetchone()[0] - 1
        #
        # cursor.execute(f'SELECT * FROM information_schema.columns WHERE table_name=\'{table}\';')
        # primary_key = cursor.fetchall()[primary_key_index]
        primary_key = list(record)[0]

    value = record[primary_key]
    if type(value) is date:
        value = f'{value:%Y%m%d}'
    elif type(value) is datetime:
        value = f'{value:%Y%m%d %H%M%S}'

    cursor.execute(f'SELECT EXISTS(SELECT 1 FROM {table} WHERE {primary_key}=%s);', [value])
    return cursor.fetchone()[0]


def database_table_is_inherited(cursor: psycopg2._psycopg.cursor, table: str) -> bool:
    """
    Whether the given table is inherited.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table

    Returns
    -------
    bool
        whether table is inherited
    """

    cursor.execute(f'SELECT EXISTS(SELECT 1 FROM pg_catalog.pg_inherits WHERE inhrelid=%s::regclass);', [f'public.{table}'])
    return cursor.fetchone()[0]


def database_table_columns(cursor: psycopg2._psycopg.cursor, table: str) -> {str: str}:
    """
    Get column names and data types of the given table, within the given database.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table

    Returns
    -------
    {str: str}
        mapping of column names to the PostGres data type
    """

    cursor.execute(f'SELECT column_name, udt_name FROM information_schema.columns WHERE table_name=%s;', [table])
    return {record[0]: record[1] for record in cursor.fetchall()}


def postgis_geometry(geometry: Union[Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon], epsg: int = None) -> str:
    """
    Convert Shapely geometry to a PostGIS geometry string.

    Parameters
    ----------
    geometry
        Shapely geometry
    epsg
        EPSG code of CRS

    Returns
    -------
    str
        PostGIS input string
    """

    if epsg is None:
        epsg = 4326

    return f'ST_SetSRID(\'{geometry.wkb_hex}\'::geometry, {epsg})'


def split_URL_port(url: str) -> (str, Union[str, None]):
    """
    Split the given URL into host and port, assuming port is appended after a colon.

    Parameters
    ----------
    url
        URL string

    Returns
    ----------
    str, Union[str, None]
        URL and port (if found)
    """

    port = None

    if url.count(':') > 0:
        url = url.split(':')
        if 'http' in url:
            url = ':'.join(url[:2])
            if len(url) > 2:
                port = int(url[2])
        else:
            url, port = url
            port = int(port)

    return url, port



import argparse
from datetime import datetime
import os
from pathlib import Path
import time

from packetraven import DEFAULT_CALLSIGNS
from packetraven.connections import APRS_fi, PacketRadio, PacketTextFile
from packetraven.tracks import APRSTrack
from packetraven.utilities import get_logger
from packetraven.writer import write_aprs_packet_tracks

LOGGER = get_logger('packetraven')

DEFAULT_INTERVAL_SECONDS = 5
DESKTOP_PATH = Path('~').expanduser() / 'Desktop'


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-k', '--key', help='API key from https://aprs.fi/page/api')
    parser.add_argument('-c', '--callsigns', help='comma-separated list of callsigns to track')
    parser.add_argument('-r', '--noradio', action='store_true', help='skip attempting to connect to APRS packet radio')
    parser.add_argument('-p', '--port', help='name of serial port connected to APRS packet radio')
    parser.add_argument('-l', '--log', help='path to log file to save log messages')
    parser.add_argument('-o', '--output', help='path to output file to save packets')
    parser.add_argument('-t', '--interval', help='seconds between each main loop')
    args = parser.parse_args()

    aprs_fi_api_key = args.key

    callsigns = args.callsigns.strip('"').split(',') if args.callsigns is not None else DEFAULT_CALLSIGNS

    if args.log is not None:
        log_filename = Path(args.log).expanduser()
        if log_filename.is_dir():
            if not log_filename.exists():
                os.makedirs(log_filename, exist_ok=True)
            log_filename = log_filename / f'{datetime.now():%Y%m%dT%H%M%S}_packetraven_log.txt'
        get_logger(LOGGER.name, log_filename)

    if args.output is not None:
        output_filename = Path(args.output).expanduser()
        output_directory = output_filename.parent
        if not output_directory.exists():
            os.makedirs(output_directory, exist_ok=True)
    else:
        output_filename = None

    serial_port = args.port

    aprs_connections = []
    if not args.noradio:
        if serial_port is not None and 'txt' in serial_port:
            try:
                text_file = PacketTextFile(serial_port)
                aprs_connections.append(text_file)
            except ConnectionError as error:
                LOGGER.warning(f'{error.__class__.__name__} - {error}')
        else:
            try:
                radio = PacketRadio(serial_port)
                serial_port = radio.location
                LOGGER.info(f'opened port {serial_port}')
                aprs_connections.append(radio)
            except ConnectionError as error:
                LOGGER.warning(f'{error.__class__.__name__} - {error}')

    try:
        aprs_api = APRS_fi(callsigns, api_key=aprs_fi_api_key)
        LOGGER.info(f'established connection to {aprs_api.location} parsing selected callsigns ({", ".join(callsigns)})')
        aprs_connections.append(aprs_api)
    except ConnectionError as error:
        LOGGER.warning(f'{error.__class__.__name__} - {error}')

    packet_tracks = {}
    while len(aprs_connections) > 0:
        LOGGER.debug(f'receiving packets from {len(aprs_connections)} source(s)')

        parsed_packets = []
        for aprs_connection in aprs_connections:
            parsed_packets.extend(aprs_connection.packets)

        LOGGER.debug(f'received {len(parsed_packets)} packets')

        if len(parsed_packets) > 0:
            for parsed_packet in parsed_packets:
                callsign = parsed_packet['callsign']

                if callsign in packet_tracks:
                    if parsed_packet not in packet_tracks[callsign]:
                        packet_tracks[callsign].append(parsed_packet)
                    else:
                        LOGGER.debug(f'{callsign:8} - received duplicate packet: {parsed_packet}')
                        continue
                else:
                    packet_tracks[callsign] = APRSTrack(callsign, [parsed_packet])
                    LOGGER.debug(f'{callsign:8} - started tracking')

                LOGGER.info(f'{callsign:8} - received new packet: {parsed_packet}')

                if 'longitude' in parsed_packet and 'latitude' in parsed_packet:
                    ascent_rate = packet_tracks[callsign].ascent_rate[-1]
                    ground_speed = packet_tracks[callsign].ground_speed[-1]
                    seconds_to_impact = packet_tracks[callsign].seconds_to_impact
                    LOGGER.info(f'{callsign:8} - Ascent rate (m/s): {ascent_rate}')
                    LOGGER.info(f'{callsign:8} - Ground speed (m/s): {ground_speed}')
                    if seconds_to_impact >= 0:
                        LOGGER.info(f'{callsign:8} - Estimated time until landing (s): {seconds_to_impact}')

            if output_filename is not None:
                write_aprs_packet_tracks(packet_tracks.values(), output_filename)

        time.sleep(DEFAULT_INTERVAL_SECONDS)


if __name__ == '__main__':
    main()



"""
Read serial connection from radio and extract APRS packets.

__authors__ = []
"""

from abc import ABC, abstractmethod
from datetime import datetime
import logging
from os import PathLike
from pathlib import Path
from typing import Any, Union

import requests
from serial import Serial
from serial.tools import list_ports
from shapely.geometry import Point

from packetraven.database import DatabaseTable
from packetraven.packets import APRSLocationPacket, LocationPacket
from packetraven.utilities import CREDENTIALS_FILENAME, get_logger, read_configuration

LOGGER = get_logger('packetraven.connection')


class PacketConnection(ABC):
    location: str

    @property
    @abstractmethod
    def packets(self) -> [APRSLocationPacket]:
        """
        List the most recent packets available from this connection.

        :return: list of APRS packets
        """

        raise NotImplementedError

    @abstractmethod
    def __enter__(self):
        raise NotImplementedError

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError


class PacketRadio(PacketConnection):
    def __init__(self, serial_port: str = None):
        """
        Connect to radio over given serial port.

        :param serial_port: port name
        """

        if serial_port is None:
            try:
                self.location = next_available_port()
            except ConnectionError:
                raise ConnectionError('could not find radio over serial connection')
        else:
            self.location = serial_port.strip('"')

        self.connection = Serial(self.location, baudrate=9600, timeout=1)

    @property
    def packets(self) -> [APRSLocationPacket]:
        return [parse_packet(line) for line in self.connection.readlines()]

    def __enter__(self):
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.connection.close()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.location}")'


class PacketTextFile(PacketConnection):
    def __init__(self, filename: PathLike = None):
        """
        Read APRS packets from a given text file.

        :param filename: path to text file, where each line consists of the time sent (`%Y-%m-%d %H:%M:%S`) followed by the raw APRS string
        """

        if not isinstance(filename, Path):
            if isinstance(filename, str):
                filename = filename.strip('"')
            filename = Path(filename)

        self.location = filename

        # open text file
        self.connection = open(self.location)

    @property
    def packets(self) -> [APRSLocationPacket]:
        return [parse_packet(line[25:].strip('\n'), datetime.strptime(line[:19], '%Y-%m-%d %H:%M:%S'))
                for line in self.connection.readlines() if len(line) > 0]

    def __enter__(self):
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.connection.close()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.location}")'


class APRS_fi(PacketConnection):
    location = 'https://api.aprs.fi/api/get'

    def __init__(self, callsigns: [str], api_key: str = None):
        """
        Connect to https://aprs.fi over given serial port.

        :param callsigns: list of callsigns to watch
        :param api_key: API key for aprs.fi
        """

        if api_key is None or api_key == '':
            configuration = read_configuration(CREDENTIALS_FILENAME)

            if 'APRS_FI' in configuration:
                api_key = configuration['APRS_FI']['api_key']
            else:
                raise ConnectionError(f'no APRS.fi API key specified')

        self.api_key = api_key
        self.callsigns = callsigns

        if not self.has_network_connection():
            raise ConnectionError(f'no network connection')

    @property
    def packets(self) -> [APRSLocationPacket]:
        query = {
            'name'  : ','.join(self.callsigns),
            'what'  : 'loc',
            'apikey': self.api_key,
            'format': 'json'
        }

        query = '&'.join(f'{key}={value}' for key, value in query.items())

        response = requests.get(f'{self.location}?{query}').json()
        if response['result'] != 'fail':
            packets = [parse_packet(packet_candidate) for packet_candidate in response['entries']]
        else:
            logging.warning(f'query failure "{response["code"]}: {response["description"]}"')
            packets = []

        return packets

    def has_network_connection(self) -> bool:
        """
        test network connection

        :return: whether current session has a network connection to APRS API
        """

        try:
            requests.get(self.location, timeout=2)
            return True
        except ConnectionError:
            return False

    def __enter__(self):
        if not self.has_network_connection():
            raise ConnectionError(f'No network connection.')

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        pass

    def __repr__(self):
        callsigns_string = ','.join(f'"{callsign}"' for callsign in self.callsigns)
        return f'{self.__class__.__name__}([{callsigns_string}], r"{self.api_key}")'


class PacketDatabaseTable(PacketConnection, DatabaseTable):
    __default_fields = {
        'time' : datetime,
        'x'    : float,
        'y'    : float,
        'z'    : float,
        'point': Point
    }

    def __init__(self, hostname: str, database: str, table: str, fields: {str: type} = None, **kwargs):
        if fields is None:
            fields = {}
        fields = {**self.__default_fields, **fields}
        super().__init__(hostname, database, table, fields, 'time', **kwargs)
        self.location = f'{self.hostname}:{self.port}/{self.database}/{self.table}'

    @property
    def packets(self) -> [LocationPacket]:
        return [LocationPacket(**{key: value for key, value in record.items() if key != 'point'}) for record in self.records]

    def __getitem__(self, time: datetime) -> LocationPacket:
        record = super().__getitem__(time)
        return LocationPacket(record['time'], record['x'], record['y'], record['z'], record['crs'])

    def __setitem__(self, time: datetime, packet: LocationPacket):
        record = {
            'time' : packet.time,
            'x'    : packet.coordinates[0],
            'y'    : packet.coordinates[1],
            'z'    : packet.coordinates[2],
            'point': Point(*packet.coordinates)
        }
        super().__setitem__(time, record)

    def __enter__(self):
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.connection.close()


class APRSPacketDatabaseTable(PacketDatabaseTable):
    __default_fields = {
        'time'    : datetime,
        'callsign': str,
        'x'       : float,
        'y'       : float,
        'z'       : float,
        'point'   : Point
    }

    def __init__(self, hostname: str, database: str, table: str, fields: {str, type} = None, **kwargs):
        if fields is None:
            fields = {}
        else:
            fields = {f'packet_{field}': field_type for field, field_type in fields.items()}
        fields = {**self.__default_fields, **fields}
        super().__init__(hostname, database, table, fields, **kwargs)

    @property
    def packets(self) -> [APRSLocationPacket]:
        return [APRSLocationPacket(**{field if field in ['time', 'x', 'y', 'z'] else field.replace('packet_', ''): value
                                      for field, value in record.items() if field not in ['point']}) for record in self.records]

    def __getitem__(self, time: datetime) -> APRSLocationPacket:
        packet = super().__getitem__(time)
        return APRSLocationPacket(packet.time, *packet.coordinates, packet.crs)

    def __setitem__(self, time: datetime, packet: APRSLocationPacket):
        record = {
            'time'    : packet.time,
            'callsign': packet.callsign,
            'x'       : packet.coordinates[0],
            'y'       : packet.coordinates[1],
            'z'       : packet.coordinates[2],
            'point'   : Point(*packet.coordinates)
        }
        super(super()).__setitem__(time, record)

    def insert(self, packets: [{str: Any}]):
        records = [{
            'time' : packet.time,
            'x'    : packet.coordinates[0],
            'y'    : packet.coordinates[1],
            'z'    : packet.coordinates[2],
            'point': Point(*packet.coordinates),
            **{f'packet_{field}': value for field, value in packet.parsed_packet.items()}
        } for packet in packets]
        super(self.__class__, self).insert(records)


def available_ports() -> str:
    """
    Iterate over available serial ports.

    :return: port name
    """

    for com_port in list_ports.comports():
        yield com_port.device
    else:
        return None


def next_available_port() -> str:
    """
    Get next port in ports list.

    :return: port name
    """

    try:
        return next(available_ports())
    except StopIteration:
        raise ConnectionError('No open serial ports.')


def parse_packet(raw_packet: Union[str, bytes, dict], packet_time: datetime = None) -> APRSLocationPacket:
    try:
        return APRSLocationPacket.from_raw_aprs(raw_packet, packet_time)
    except Exception as error:
        logging.exception(f'{error.__class__.__name__} - {error} for raw packet "{raw_packet}"')



#!/usr/bin/env python
from dunamai import Version
from setuptools import config, find_packages, setup

metadata = config.read_configuration('setup.cfg')['metadata']

setup(
    name=metadata['name'],
    version=Version.from_any_vcs().serialize(),
    author=metadata['author'],
    author_email=metadata['author_email'],
    description=metadata['description'],
    long_description=metadata['long_description'],
    long_description_content_type="text/markdown",
    url=metadata['url'],
    packages=find_packages(),
    python_requires='>=3.6',
    setup_requires=['dunamai', 'setuptools>=41.2', 'wheel'],
    install_requires=[
        'aprslib',
        'haversine',
        'numpy',
        'pyserial',
        'geojson',
        'fastkml',
        'requests',
        'pyproj',
        'psycopg2',
        'shapely',
        'sshtunnel'
    ],
    extras_require={'dev': ['coverage', 'flake8', 'nose']},
    entry_points={
        'console_scripts': ['packetraven=client.packetraven_cli:main'],
        'gui_scripts'    : ['packetraven_gui=client.packetraven_gui:main']
    },
    test_suite='nose.collector'
)
